/*
SPDX-FileCopyrightText: Simon Schneegans <code@simonschneegans.de>
SPDX-License-Identifier: MIT
*/

/* This is used in the tutorial. */
.tutorial-video {
    box-shadow: 2px 2px 3px rgba(0, 0, 0, 0.3);
}

.pill-button {
    border-radius: 9999px;
}

.round-button {
    border-radius: 20px;
}

.small-popover>contents {
    padding: 0;
}


/* This is used for the unread-achievements counter. */
.counter-label {
    text-shadow: none;
    color: @theme_selected_fg_color;
    background-color: @theme_selected_bg_color;
    font-size: smaller;
    border-radius: 5px;
    padding: 0px 8px;
}

/* This is an entry at the top of the menu editor. */
.menu-editor-path-item {
    border-radius: 0;
    border-left: none;
    border-top: none;
    border-bottom: none;
}

.menu-editor-path-item:first-child {
    border-radius: 8px 0 0 8px;
}

menubutton.menu-editor-add-button button {
    border-radius: 0 8px 8px 0;
    border-right: none;
    border-top: none;
    border-bottom: none;
}

/* Remove the padding inside the Adw.Clamp in the settings dialog. */
preferencespage clamp>box {
    margin: 0;
}